#version 330

const int MAX_POINT_LIGHTS = 2;                                                                               
                                                                                    
struct PointLight                                                                           
{              
    vec3 Color;                                                                     
    float AmbientInt;                                                         
    float DiffuseInt;                                                                                                                                                                      
    vec3 Position;                                                                          
    float ConstAtt;                                                              
    float LinearAtt;                                                                   
    float ExpAtt;                                                                           
};                                                                                          
                                                                                            
uniform int numPointLights;                                                                                                               
uniform PointLight pointLights[MAX_POINT_LIGHTS];    

uniform sampler2D colorTexture;
uniform sampler2DShadow shadowMap;

uniform sampler2D normalMap;
uniform sampler2D depthMap;
uniform float heightScale;

uniform vec3 sunLightDir;
uniform vec3 sunLightCol;
uniform float sunLightInt;
uniform vec3 ambLightCol;
uniform float ambLightInt;
uniform float specLightInt;
uniform float shininess;

uniform  vec4 fogColor; // Fog color
uniform float fogStart; // This is only for linear fog
uniform float fogEnd; // This is only for linear fog
uniform float fogDensity; // For exp and exp2 equation 
uniform int fogEquation; // 0 = linear, 1 = exp, 2 = exp2

// omni shadowmap
uniform samplerCube cm_z_tex;
uniform vec4 l_pos; // world space light position
uniform vec2 near_far; // near and far plane for cm-cams


in vec2 fUVs;
in vec3 N; //normal
in vec3 V; //view
in vec3 P; //position
in vec3 T; //tangent
in vec3 B; //bitangent

//in vec3 tangentLightPos_Dir;
//in vec3 tangentLightPos_Point[MAX_POINT_LIGHTS];
//in vec3 tangentViewPos;
//in vec3 tangentFragPos;

in vec3 vertViewDist; //distance view to vertex

in vec4 SM_tex_coord;

out vec4 fragColor;

vec3 sampleOffsetDirections[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);   


float getFogFactor(int fogEquation, float fogStart, float fogEnd, float fogDensity, float fogCoord)
{
   float fResult = 0.0;
   if(fogEquation == 0)
      fResult = (fogEnd-fogCoord)/(fogEnd - fogStart);
   else if(fogEquation == 1)
      fResult = exp(-fogDensity*fogCoord);
   else if(fogEquation == 2)
      fResult = exp(-pow(fogDensity*fogCoord, 2.0));
      
   fResult = 1.0-clamp(fResult, 0.0, 1.0);

   return fResult;
}







vec3 CalcPointLight(int index, vec3 N_, vec3 V_, float shadowOmni)
{
	PointLight light = pointLights[index];

	vec3 L = light.Position - P;
	float Distance = length(L);

	vec3 H = normalize(V_+normalize(L));
	float cosNL = max(dot(N_,normalize(L)),0.0f);
	
	vec3 ambColor = light.Color*light.AmbientInt;
	vec3 lightColor = light.Color*((1.0f-light.AmbientInt)*light.DiffuseInt);
	vec3 specColor = light.Color*specLightInt;
	float dif_factor = cosNL;
	float spec_factor = cosNL * pow(max(dot(N_, H), 0),shininess);

	vec3 dif = lightColor * dif_factor * (1.0 - shadowOmni); // * visibility;
	vec3 spec = specColor*spec_factor * (1.0 - shadowOmni); // * visibility;
    vec3 amb = ambLightCol*ambLightInt; //ambLightCol*ambLightInt;

	vec3 Color = dif+spec+amb;	
	//vec3 Color = dif;

    float Attenuation = pointLights[index].ConstAtt +
                        pointLights[index].LinearAtt * Distance +
                        pointLights[index].ExpAtt * Distance * Distance;

    return Color / Attenuation;
} 

/**
* Omni Directional Shadowmapping with percentage-closer filtering.
* The function uses two global variables: P (fragment position), l_pos (light position of the pointlight) and sampleOffsetDirections for PCF as "filter kernel".
*/
float shadowOmniCalculation(){

	float shadow = 0.0;
	float bias = 0.15;
	int samples = 20;
	//float viewDistance = length(vertViewDist - P);
	//float diskRadius = (1.0 + (viewDistance / near_far.y)) / 25.0; 
	float diskRadius = 0.10;
	vec3 fragToLight = P - l_pos.xyz;
	float currentDepth = length(fragToLight);
	for(int i = 0; i < samples; ++i)
	{
		float closestDepth = texture(cm_z_tex, fragToLight + sampleOffsetDirections[i] * diskRadius).r;
		closestDepth = closestDepth * (near_far.y - near_far.x) + near_far.x;
		if(currentDepth - bias > closestDepth)
			shadow += 0.85;
	}
	shadow /= float(samples);  

	return shadow;

	/*
	omnidirectional shadowmapping without PCF

	float shadowOmni = 0.0;
	// get direction vector between fragment position and light position
	vec3 fragToLight = P - l_pos.xyz;
	// use the light to fragment vector to sample from the depth map
	float closestDepth = texture(cm_z_tex, fragToLight).r;
	// currently range [0,1], re transform back to original value.
	float closestDepth2 = closestDepth;
	closestDepth = closestDepth * (near_far.y - near_far.x) + near_far.x;

	// now get current linear depth as the lenght between fragment and light position
	float currentDepth = length(fragToLight);
	// to avoid shadow acne
	float biass = 0.15;
	
	if(closestDepth + biass < currentDepth){
		//shadow
		shadowOmni = 1.0;
	}else{
		// no shadow
		shadowOmni = 0.15;
	}
	return shadowOmni;
	*/
}
 
void main() {
	float visibility = 1.0;
	float shadowOmni = 1.0;

	vec3 C;
	vec3 L;
	vec3 V_;
	vec3 N_;
	vec4 texColor;
	vec2 texCoords = fUVs;

	
	// Normal Mapping
	bool parallax = false;
	if (heightScale > 0.0f) {
		parallax = true;
	}

	if (parallax) {
		vec3 normal = normalize((texture2D( normalMap, texCoords).rgb * 2.0 -1.0) );// - vec3(1.0, 1.0, 1.0));
		normal.x = normal.x * -1.0;

		mat3 TBN = mat3(normalize(B), normalize(T), normalize(N));
		N_ = TBN * normal;
		N_ = normalize(N_);

	} else {
		N_ = normalize(N);
	}
	
	


	texColor = texture(colorTexture, texCoords);
	C = texColor.rgb;
	
	/*
	L = normalize(-(sunLightDir));
	V_ = normalize(V);

	// Directional Light Computation ___________________________________________
	vec3 H = normalize(V_ + L);
	float cosNL = max(dot(N_,L),0);
	vec3 sunColor = sunLightCol*((1-ambLightInt)*sunLightInt);
	vec3 specColor = sunLightCol*specLightInt;
	float dif_factor = cosNL;
	float spec_factor = cosNL * pow(max(dot(N_, H), 0),shininess);
	// Directional Light Computation ___________________________________________

	// Directional Shadow Mapping ___________________________________________
	//--------------------------------------------------
/*  float cosNL2 = max( dot(normalize(N), L ),0);
	float bias = 0.005*tan(acos(cosNL2));
	vec4 SMtex = SM_tex_coord;
	SMtex.z += bias;
	visibility = min(textureProj(shadowMap, SMtex)*2.0f ,1.0f); //0.5;
*/

	// Omni directional shadow map cube ___________________________________________
	shadowOmni = shadowOmniCalculation();
	// Test for debug
	//fragColor = vec4(vec3(shadowOmni), 1.0);
	// Omni directional shadow map cube ___________________________________________
	
	/*
	vec3 dif = sunColor * dif_factor; // * visibility
	vec3 spec = specColor * spec_factor; // * visibility
    vec3 amb = ambLightCol*ambLightInt; // * visibility
	*/
	vec3 point = vec3(0.0, 0.0, 0.0);

	
	for (int i = 0 ; i < numPointLights ; i++) {
         point += CalcPointLight(i, N_, V, shadowOmni);
    }
	//point = CalcPointLightDefault();
	
	fragColor = vec4(C *(point), 1.0); //spec+amb+dif of directional light
	float fogCoord = length(vertViewDist); 
	float fogfactor = getFogFactor(fogEquation, fogStart, fogEnd, fogDensity, fogCoord);
	fragColor = mix(fragColor, fogColor, fogfactor);
	fragColor = vec4(fragColor.rgb, texColor.a);
	
	
}